"""
Converters and structure hooks for various data types.
"""

import dataclasses
import re
from datetime import date, datetime, timedelta
from enum import Enum
from pathlib import Path
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from ._compat import PY_310, PY_311


if PY_310:
    from types import UnionType
else:
    from typing import Union as UnionType  # type: ignore

from .types import ET, T


if TYPE_CHECKING:
    import cattrs
    import pydantic
#
#
# __all__ = [
#     "Converter",
#     "TSConverter",
#     "default_converter",
#     "get_default_structure_hooks",
#     "register_attrs_hook_factory",
#     "register_strlist_hook",
#     "to_datetime",
#     "to_bool",
#     "to_enum",
#     "to_path",
#     "to_resolved_path",
# ]


class Converter(Protocol):
    """
    **Protocol** that converters must implement.

    Only a :meth:`structure()` method similar to the one from :program:`cattrs` is
    required.

    .. versionadded:: 23.1.0
    """

    def structure(self, value: Any, cls: Type[T]) -> T:
        """
        Convert *value* to an instance of *cls* and return it.

        Args:
            value: The data to be converted.
            cls: The type to convert *value* to.

        Return:
            An instance of *cls* for *value*.
        """
        ...


class TSConverter:
    """
    A simple converter that can replace :program:`cattrs` if you want to use
    Typed Settings without dependencies.

    It supports the same types as the default :program:`cattrs` converter.
    """

    def __init__(
        self,
        resolve_paths: bool = True,
        strlist_sep: Union[str, Callable[[str], list], None] = ":",
    ) -> None:
        if strlist_sep is None:
            self.strlist_hook: Optional[Callable[[str], list]] = None
        elif isinstance(strlist_sep, str):
            self.strlist_hook = lambda v: v.split(strlist_sep)  # type: ignore
        else:
            self.strlist_hook = strlist_sep

        self.scalar_converters: Dict[Any, Callable[[Any, type], Any]] = {
            Any: to_any,
            bool: to_bool,
            int: to_type,
            float: to_type,
            str: to_type,
            datetime: to_datetime,
            date: to_date,  # Must come after "datetime" b/c of subclassing!
            timedelta: to_timedelta,
            Enum: to_enum,
            Path: to_resolved_path if resolve_paths else to_path,
        }
        try:
            import pydantic

            self.scalar_converters[pydantic.SecretBytes] = to_pydantic_secretbytes
            self.scalar_converters[pydantic.SecretStr] = to_pydantic_secretstr
        except ImportError:
            pass

        self.composite_hook_factories: List[HookFactory] = [
            ListHookFactory,
            TupleHookFactory,
            DictHookFactory,
            MappingProxyTypeHookFactory,
            SetHookFactory,
            FrozenSetHookFactory,
            UnionHookFactory,
            AttrsHookFactory,
            DataclassesHookFactory,
            PydanticHookFactory,
        ]

    def structure(self, value: Any, cls: Type[T]) -> T:
        """
        Convert *value* to an instance of *cls* and return it.

        Args:
            value: The data to be converted.
            cls: The type to convert *value* to.

        Return:
            An instance of *cls* for *value*.
        """
        for ctype, convert in self.scalar_converters.items():
            if cls is ctype or (
                ctype is not Any and isinstance(cls, type) and issubclass(cls, ctype)
            ):
                return convert(value, cls)

        origin = get_origin(cls)
        args = get_args(cls)
        for hook in self.composite_hook_factories:
            if hook.match(cls, origin, args):
                convert = hook.get_structure_hook(self, cls, origin, args)
                return convert(value, cls)

        raise TypeError(f"Cannot create converter for generic type: {cls}")

    def maybe_apply_strlist_hook(self, value: T) -> Union[list, T]:
        """
        Apply the string list hook to *value* if one is defined and if *value* is a
        string.
        """
        if self.strlist_hook and isinstance(value, str):
            return self.strlist_hook(value)
        return value


def default_converter(*, resolve_paths: bool = True) -> Converter:
    """
    Get a default instances of a converter which will be used to convert/structured
    the loaded settings.

    Args:
        resolve_paths: Whether or not to resolve relative paths.

    Return:
        If :program:`cattrs` is installed, a :class:`cattrs.Converter`.  Else, a
        :class:`TSConverter`.  The converters are configured to handle the following
        types:

        - :class:`bool` (see :func:`to_bool()` for supported inputs)
        - :class:`int`
        - :class:`float`
        - :class:`str`
        - :class:`datetime.datetime` (see :func:`to_datetime()`)
        - :class:`datetime.date` (see :func:`to_date()`)
        - :class:`datetime.timedelta` (see :func:`to_timedelta()`)
        - :class:`enum.Enum` using (see :func:`to_enum()`)
        - :class:`pathlib.Path` (see :func:`to_path()` and :func:`to_resolved_path()`)
        - :class:`list`
        - :class:`tuple`
        - :class:`dict`
        - :class:`types.MappingProxyType` ("read-only" dicts)
        - :class:`set`
        - :class:`frozenset`
        - :data:`typing.Optional`
        - :data:`typing.Union` (depending on the converter, only to a certain degree,
          but this should not be relevant for settings with clearly defined types)
        - :mod:`attrs` classes (from instances and dicts)

        :class:`list`, :class:`tuple`, :class:`set`, and :class:`frozenset` set can also
        be converted from strings.  By default, strings are split by ``:``.  See
        :class:`TSConverter` or :func:`register_strlist_hook()` for details.

    This converter can also be used as a base for converters with custom
    structure hooks.

    .. versionchanged:: 23.1.0
       Return a :program:`cattrs` converter if it is installed or else a Typed Settings
       converter.
    .. versionchanged:: 24.3.0
       Added :func:`to_date()` and :func:`to_timedelta()`.
    """
    try:
        import cattrs  # noqa: F401
    except ImportError:
        return get_default_ts_converter(resolve_paths=resolve_paths)
    else:
        return get_default_cattrs_converter(resolve_paths=resolve_paths)


def get_default_ts_converter(resolve_paths: bool = True) -> "TSConverter":
    """
    Return a :class:`TSConverter` with default settings
    (see :func:`default_converter()` for argument and return value description).

    Args:
        resolve_paths: Whether or not to resolve relative paths.

    Return:
        A :class:`TSConverter` instance with default configuration.
    """
    return TSConverter(resolve_paths=resolve_paths)


def get_default_cattrs_converter(resolve_paths: bool = True) -> "cattrs.Converter":
    """
    Return a :class:`cattrs.Converter` with default settings
    (see :func:`default_converter()` for argument and return value description).

    Args:
        resolve_paths: Whether or not to resolve relative paths.

    Return:
        A :class:`cattrs.Converter` instance with default configuration.

    Raises:
        ModuleNotFoundError: if :program:`cattrs` is not installed.
    """
    try:
        import cattrs
    except ImportError as e:
        raise ModuleNotFoundError(
            "Module 'cattrs' not installed.  Please run "
            "'python -m pip install -U typed-settings[cattrs]'"
        ) from e

    converter = cattrs.Converter()
    register_mappingproxy_hook(converter)
    register_attrs_hook_factory(converter)
    register_dataclasses_hook_factory(converter)
    register_pydantic_hook_factory(converter)
    register_strlist_hook(converter, ":")
    for t, h in get_default_structure_hooks(resolve_paths=resolve_paths):
        converter.register_structure_hook(t, h)  # type: ignore
    return converter


def get_default_structure_hooks(
    *,
    resolve_paths: bool = True,
) -> List[Tuple[type, Callable[[Any, type], Any]]]:
    """
    Return a list of default structure hooks for cattrs.

    Args:
        resolve_paths: Whether or not to resolve relative paths.

    Return:
        A list of tuples that can be used as args for
        :meth:`cattrs.BaseConverter.register_structure_hook()`.
    """
    path_hook = to_resolved_path if resolve_paths else to_path
    hooks: List[Tuple[type, Callable[[Any, type], Any]]] = [
        (bool, to_bool),
        (datetime, to_datetime),
        (date, to_date),
        (timedelta, to_timedelta),
        (Enum, to_enum),
        (Path, path_hook),
    ]
    try:
        import pydantic

        hooks.append((pydantic.SecretBytes, to_pydantic_secretbytes))
        hooks.append((pydantic.SecretStr, to_pydantic_secretstr))
    except ImportError:
        pass
    return hooks


def register_attrs_hook_factory(converter: "cattrs.Converter") -> None:
    """
    Register a hook factory that allows using instances of :program:`attrs` classes
    where :program:`cattrs` would normally expect a dictionary.

    These instances are then returned as-is and without further processing.

    Args:
        converter: The :class:`cattrs.Converter` to register the hook at.
    """

    def allow_attrs_instances(typ):  # type: ignore[no-untyped-def]
        def structure_attrs(val, _):  # type: ignore[no-untyped-def]
            if isinstance(val, typ):
                return val
            return converter.structure_attrs_fromdict(val, typ)

        return structure_attrs

    import attrs

    converter.register_structure_hook_factory(attrs.has, allow_attrs_instances)


def register_dataclasses_hook_factory(converter: "cattrs.Converter") -> None:
    """
    Register a hook factory that allows using instances of :mod:`dataclasses` classes
    where :program:`cattrs` would normally expect a dictionary.

    These instances are then returned as-is and without further processing.

    Args:
        converter: The :class:`cattrs.Converter` to register the hook at.
    """

    def allow_dataclasses_instances(typ):  # type: ignore[no-untyped-def]
        def structure_dataclasses(val, _):  # type: ignore[no-untyped-def]
            if isinstance(val, typ):
                return val
            return converter.structure_attrs_fromdict(val, typ)

        return structure_dataclasses

    converter.register_structure_hook_factory(
        dataclasses.is_dataclass, allow_dataclasses_instances
    )


def register_pydantic_hook_factory(converter: "cattrs.Converter") -> None:
    """
    Register a hook factory that allows using instances of :mod:`dataclasses` classes
    where :program:`cattrs` would normally expect a dictionary.

    These instances are then returned as-is and without further processing.

    Args:
        converter: The :class:`cattrs.Converter` to register the hook at.
    """
    try:
        import pydantic
    except ImportError:  # pragma: no cover
        return

    def check(typ: type) -> bool:
        return issubclass(typ, pydantic.BaseModel)

    def to_pydantic(typ):  # type: ignore[no-untyped-def]
        def structure_dataclasses(val, _):  # type: ignore[no-untyped-def]
            if isinstance(val, typ):
                return val
            return typ(**val)

        return structure_dataclasses

    converter.register_structure_hook_factory(check, to_pydantic)


def register_mappingproxy_hook(converter: "cattrs.Converter") -> None:
    """
    Register a hook factory for converting data to :class:`types.MappingProxyType`
    instances.

    Args:
        converter: The :class:`cattrs.Converter` to register the hook at.
    """

    def check(cls: type) -> bool:
        return cls is MappingProxyType or get_origin(cls) is MappingProxyType

    def convert(val: Mapping, cls: Type[T]) -> T:
        args = get_args(cls)
        t = Dict[args[0], args[1]] if args else Dict  # type: ignore
        return MappingProxyType(converter.structure(val, t))  # type: ignore

    converter.register_structure_hook_func(check, convert)


def register_strlist_hook(
    converter: "cattrs.Converter",
    sep: Optional[str] = None,
    fn: Optional[Callable[[str], list]] = None,
) -> None:
    """
    Register a hook factory with *converter* that allows structuring lists,
    (frozen) sets and tuples from strings (which may, e.g., come from
    environment variables).

    Args:
        converter: The :class:`cattrs.Converter` to register the hook at.
        sep: A separator used for splitting strings (see :meth:`str.split()`).
            Cannot be used together with *fn*.
        fn: A function that takes a string and returns a list, e.g.,
            :func:`json.loads()`.  Cannot be used together with *sep*.

    Example:
        >>> from typing import List
        >>>
        >>> converter = default_converter()
        >>> register_strlist_hook(converter, sep=":")
        >>> converter.structure("1:2:3", List[int])
        [1, 2, 3]
        >>>
        >>> import json
        >>>
        >>> converter = default_converter()
        >>> register_strlist_hook(converter, fn=json.loads)
        >>> converter.structure("[1,2,3]", List[int])
        [1, 2, 3]

    """
    if (sep is None and fn is None) or (sep is not None and fn is not None):
        raise ValueError('You may either pass "sep" *or* "fn"')
    if sep is not None:
        fn = lambda v: v.split(sep)  # noqa

    from cattrs._compat import is_tuple
    from cattrs.cols import is_frozenset, is_sequence, is_set, list_structure_factory

    collection_types = [
        # Order is important, tuple must be last!
        (is_sequence, list_structure_factory),
        (is_set, lambda _, converter: converter._structure_set),
        (is_frozenset, lambda _, converter: converter._structure_frozenset),
        (is_tuple, lambda _, converter: converter._structure_tuple),
    ]

    for check, structure_func_factory in collection_types:
        hook_factory = _generate_hook_factory(structure_func_factory, fn)
        converter.register_structure_hook_factory(check, hook_factory)


def _generate_hook_factory(structure_func_factory, fn):  # type: ignore[no-untyped-def]
    def gen_func(typ, converter):  # type: ignore[no-untyped-def]
        base_hook = structure_func_factory(typ, converter)
        def str2collection(val, _, base_hook=base_hook):  # type: ignore[no-untyped-def]
            if isinstance(val, str):
                val = fn(val)
            return base_hook(val, typ)

        return str2collection

    return gen_func


def to_any(value: Any, _cls: type) -> Any:
    """
    Return *value* as-is.
    """
    return value


def to_bool(value: Any, _cls: type = bool) -> bool:
    """
    Convert "boolean" strings (e.g., from env. vars.) to real booleans.

    Values mapping to :code:`True`:

    - :code:`True`
    - :code:`"true"` / :code:`"t"` (case insensitive)
    - :code:`"yes"` / :code:`"y"` (case insensitive)
    - :code:`"on"` (case insensitive)
    - :code:`"1"`
    - :code:`1`

    Values mapping to :code:`False`:

    - :code:`False`
    - :code:`"false"` / :code:`"f"` (case insensitive)
    - :code:`"no"` / :code:`"n"` (case insensitive)
    - :code:`"off"` (case insensitive)
    - :code:`"0"`
    - :code:`0`

    Args:
        value: The value to parse.
        _cls: (ignored)

    Return:
        A :class:`bool` for the input *value*.

    Raise:
        ValueError: If *value* is any other value than stated above.
    """
    if isinstance(value, str):
        value = value.lower()
    truthy = {True, "true", "t", "yes", "y", "on", "1", 1}
    falsy = {False, "false", "f", "no", "n", "off", "0", 0}
    try:
        if value in truthy:
            return True
        if value in falsy:
            return False
    except TypeError:
        # Raised when "val" is not hashable (e.g., lists)
        pass
    raise ValueError(f"Cannot convert value to bool: {value}")


def to_date(value: Union[datetime, str], cls: Type[date] = date) -> date:
    """
    Convert an ISO formatted string to :class:`datetime.date`.  Leave the input
    untouched if it is already a date.

    See: :meth:`datetime.date.fromisoformat()`

    Args:
        value: The input data
        cls: The target type.  Must be :class:`datetime.date` or a subclass.

    Return:
        The converted date instance

    Raise:
        TypeError: If *value* is neither a string nor a date
        ValueError: If *value* cannot be parsed as date.

    .. versionadded:: 24.3.0
    """
    if not isinstance(value, (date, str)):
        raise TypeError(
            f"Invalid type {type(value).__name__!r}; expected 'date' or " f"'str'."
        )
    if isinstance(value, str):
        return cls.fromisoformat(value)
    return value


def to_datetime(
    value: Union[datetime, str], cls: Type[datetime] = datetime
) -> datetime:
    """
    Convert an ISO formatted string to :class:`datetime.datetime`.  Leave the input
    untouched if it is already a datetime.

    See: :meth:`datetime.datetime.fromisoformat()`

    The ``Z`` suffix is supported on Python versions prior to 3.11, too.

    Args:
        value: The input data
        cls: The target type.  Must be :class:`datetime.datetime` or a subclass.

    Return:
        The converted datetime instance

    Raise:
        TypeError: If *value* is neither a string nor a datetime
        ValueError: If *value* cannot be parsed as datetime.
    """
    if not isinstance(value, (datetime, str)):
        raise TypeError(
            f"Invalid type {type(value).__name__!r}; expected 'datetime' or " f"'str'."
        )
    if isinstance(value, str):
        if not PY_311 and value[-1] == "Z":
            value = value.replace("Z", "+00:00")
        return cls.fromisoformat(value)
    return value


_SIGN = "(?P<sign>[+-])?"
_DAYS = "(?P<days>[0-9]+)"
_HOURS = "(?P<hours>[0-9]+)"
_MINUTES = "(?P<minutes>[0-9]+)"
_SECONDS = r"(?P<seconds>[0-9]+)(\.(?P<micros>[0-9]{1,6}))?"
# [±][{D}d[,]][[{HH}:]{MM}:]{SS}[.{ffffff}]
RE_TIMEDELTA_SIMPLE = re.compile(
    f"^{_SIGN}({_DAYS}D,?)?(({_HOURS}:)?{_MINUTES}:)?{_SECONDS}$", flags=re.IGNORECASE
)
# [±][{D}d][{HH}h][{MM}m][{SS}[.{ffffff}]s]
RE_TIMEDELTA_SIMPLE_ISO = re.compile(
    f"^{_SIGN}({_DAYS}D)?({_HOURS}H)?({_MINUTES}M)?({_SECONDS}S)?$", flags=re.IGNORECASE
)
# [±]P[{D}D][T[{HH}H][{MM}M][{SS}[.{ffffff}]S]]
RE_TIMEDELTA_ISO = re.compile(
    f"^{_SIGN}"
    r"P(?!\b)"  # "P", but not on a word boundary (e.g., at the end of the string)
    f"({_DAYS}D)?"
    f"("
    r"T(?!\b)"  # "T", but not on a word boundary (e.g., at the end of the string)
    f"({_HOURS}H)?"
    f"({_MINUTES}M)?"
    f"({_SECONDS}S)?"
    r")?$",
    flags=re.IGNORECASE,
)


def to_timedelta(
    value: Union[timedelta, int, float, str], cls: Type[timedelta]
) -> timedelta:
    """
    Convert *value* to a :class:`datetime.timedelta`.

    Accepts strings, integers and floats, and timedelta instances.

    Timedelta instances are returned unchanged.

    Integers and floats are interpreted as seconds.

    Supported string formats (all are case-insensitive):

    - :samp:`[±]P[{D}D][T[{HH}H][{MM}M][{SS}[.{ffffff}]S]]` (`ISO durations`_), e.g.:

      - ``P1DT03H04M05S``
      - ``-P180D``
      - ``PT4H30M``
      - ``P1DT30S``

    - :samp:`[±][{D}d][{HH}h][{MM}m][{SS}[.{ffffff}]s]` (simplified ISO variant), e.g.:

      - ``1d3h4m5s``
      - ``-180d``
      - ``4h30m``
      - ``1d30s``

    - :samp:`[±][{D}d[,]][[{HH}:]{MM}:]{SS}[.{ffffff}]`, e.g.:

      - ``1d,03:04:05``
      - ``-180D``
      - ``4:30:00``
      - ``1d30``

    .. _iso durations: https://en.wikipedia.org/wiki/ISO_8601#Durations

    Args:
        value: The input data
        cls: The target type.  Must be :class:`datetime.timedelta` or a subclass.

    Return:
        The converted timedelta instance

    Raise:
        TypeError: If *value* is neither a string, float, or int nor a timedelta.
        ValueError: If *value* cannot be parsed as timedelta.

    .. versionadded:: 24.3.0
    """
    if not isinstance(value, (timedelta, str, float, int)):
        raise TypeError(
            f"Invalid type {type(value).__name__!r}; expected 'timedelta', 'float', "
            f"'int', or 'str'."
        )

    if isinstance(value, timedelta):
        return value

    if isinstance(value, (int, float)):
        return cls(seconds=value)

    for regex in (RE_TIMEDELTA_ISO, RE_TIMEDELTA_SIMPLE_ISO, RE_TIMEDELTA_SIMPLE):
        match = regex.match(value)
        if match:
            break
    else:
        raise ValueError(f"Cannot parse value as timedelta: {value}")

    parts = match.groupdict(default="0")
    days = int(parts["days"])
    hours = int(parts["hours"])
    minutes = int(parts["minutes"])
    seconds = int(parts["seconds"])
    # Append "0" to "micros" to get a 6-digit number (.7 -> 7 -> 700_000)
    micros = int(f"{parts['micros']:<06}")

    td = cls(
        days=days, hours=hours, minutes=minutes, seconds=seconds, microseconds=micros
    )
    if parts["sign"] == "-":
        return -td
    return td


def timedelta_to_str(td: timedelta) -> str:
    """
    Serialize a timedelta to a string that can be parsed by :func:`to_timedelta()`.
    """
    if td == abs(td):
        is_negative = False
    else:
        is_negative = True
        td = -td
    days = td.days
    hours, seconds = divmod(td.seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    micros = f".{td.microseconds:>06}".rstrip("0") if td.microseconds else ""
    result = (
        f"{f'{days}d' if days else ''}"
        f"{f'{hours}h' if hours else ''}"
        f"{f'{minutes}m' if minutes else ''}"
        f"{f'{seconds}{micros}s' if seconds or micros else ''}"
    )
    if result and is_negative:
        result = f"-{result}"
    return result


def to_enum(value: Any, cls: Type[ET]) -> ET:
    """
    Return an instance of the enum *cls* for *value*.

    If the to be converted value is not already an enum, the converter will
    create one by name (``MyEnum[val]``).

    Args:
        value: The input data
        cls: The enum type

    Return:
        An instance of *cls*

    Raise:
        KeyError: If *value* is not a valid member of *cls*
    """
    if isinstance(value, cls):
        return value

    return cls[value]


def to_path(value: Union[Path, str], _cls: type) -> Path:
    """
    Convert *value* to :class:`~pathlib.Path`.

    Args:
        value: The input data
        _cls: (ignored)

    Return:
        An instance of :class:`~pathlib.Path`

    Raise:
        TypeError: If *value* cannot be converted to a path.
    """
    return Path(value)


def to_resolved_path(value: Union[Path, str], _cls: type) -> Path:
    """
    Convert *value* to :class:`~pathlib.Path` and resolve it.

    Args:
        value: The input data
        _cls: (ignored)

    Return:
        A resolved instance of :class:`~pathlib.Path`

    Raise:
        TypeError: If *value* cannot be converted to a path.
    """
    return Path(value).resolve()


def to_type(value: Any, cls: Type[T]) -> T:
    """
    Convert *value* to *cls*.

    Args:
        value: The input data
        cls: A class that takes a single argument, e.g., :class:`int`, :class:`float`,
            or :class:`str`.

    Return:
        An instance of *cls*.

    Raise:
        ValueError: if *value* cannot be converted to *cls*.
    """
    return cls(value)  # type: ignore[call-arg]


def to_pydantic_secretbytes(
    value: Any, _cls: "Type[pydantic.SecretBytes]"
) -> "pydantic.SecretBytes":
    """
    Convert *value* to :class:`pydantic.SecretStr`.

    Args:
        value: The input data
        _cls: (ignored)

    Return:
        An instance of *cls*.
    """
    import pydantic

    if isinstance(value, pydantic.SecretBytes):
        return value

    return pydantic.SecretBytes(value)


def to_pydantic_secretstr(
    value: Any, _cls: "Type[pydantic.SecretStr]"
) -> "pydantic.SecretStr":
    """
    Convert *value* to :class:`pydantic.SecretStr`.

    Args:
        value: The input data
        cls: A pydantic SecretStr class or any subclass
        _cls: (ignored)

    Return:
        An instance of *cls*.
    """
    import pydantic

    if isinstance(value, pydantic.SecretStr):
        return value

    return pydantic.SecretStr(value)


class HookFactory(Protocol):
    """
    **Protocol** for :class:`TSConverter` hook factories.

    Hook factories have a :meth:`match` functions that decides whether they can handle
    given type/class.  In addition, they can generate a structure hook for that type.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        """
        Check whether this class can handle the given type *cls*.

        Args:
            cls: The type/class to check.
            origin: The type's origin as returned by :func:`typing.get_origin()`.
            args: The type's args as retuned by :func:`typing.get_args()`.

        Return:
            ``True`` if this class can convert the given type, or else ``False``.
        """
        ...

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Any, Type[T]], T]:
        """
        Return a structure hook for the given type/class.

        Args:
            converter: The :class:`TSConverter` that the returned hook will be
                registered at.  The structure hook can use the converter to recursively
                convert sub elements of composite types.
            cls: The type/class to convert to.
            origin: The type's origin as returned by :func:`typing.get_origin()`.
            args: The type's args as retuned by :func:`typing.get_args()`.

        Return:
            A structure hook, which is a function
            :samp:`hook({value}: Any, {cls}: Type[T]) -> T`.
        """
        ...


class AttrsHookFactory:
    """
    A :class:`HookFactory` that returns :program:`attrs` classes from dicts.  Instances
    are of the given class are accepted as well and returned as-is (without further
    processing of their attributes).
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        try:
            import attrs
        except ImportError:
            return False

        return attrs.has(cls)

    @staticmethod
    def get_structure_hook(
        converter: Converter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Union[dict, T], Type[T]], T]:
        import attrs

        def convert(value: Union[dict, T], cls: Type[T]) -> T:
            if isinstance(value, cls):
                return value

            if not isinstance(value, dict):
                raise TypeError(
                    f'Invalid type "{type(value).__name__}"; expected '
                    f'"{cls.__name__}" or "dict".'
                )

            fields = attrs.fields_dict(cls)  # type: ignore[arg-type]
            values = {
                n: converter.structure(v, fields[n].type)  # type: ignore[arg-type]
                for n, v in value.items()
            }
            return cls(**values)

        return convert


class DataclassesHookFactory:
    """
    A :class:`HookFactory` that returns :mod:`dataclasses` from dicts.  Instances
    are of the given class are accepted as well and returned as-is (without further
    processing of their attributes).
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return dataclasses.is_dataclass(cls)

    @staticmethod
    def get_structure_hook(
        converter: Converter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Union[dict, T], Type[T]], T]:
        def convert(value: Union[dict, T], cls: Type[T]) -> T:
            if isinstance(value, cls):
                return value

            if not isinstance(value, dict):
                raise TypeError(
                    f'Invalid type "{type(value).__name__}"; expected '
                    f'"{cls.__name__}" or "dict".'
                )

            fields = {f.name: f for f in dataclasses.fields(cls)}  # type: ignore[arg-type]
            values = {
                n: converter.structure(v, fields[n].type)  # type: ignore[arg-type]
                for n, v in value.items()
            }
            return cls(**values)

        return convert


class PydanticHookFactory:
    """
    A :class:`HookFactory` that returns :mod:`dataclasses` from dicts.  Instances
    are of the given class are accepted as well and returned as-is (without further
    processing of their attributes).
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        try:
            import pydantic
        except ImportError:
            return False
        return issubclass(cls, pydantic.BaseModel)

    @staticmethod
    def get_structure_hook(
        converter: Converter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Union[dict, T], Type[T]], T]:
        def convert(value: Union[dict, T], cls: Type[T]) -> T:
            if isinstance(value, cls):
                return value

            if not isinstance(value, dict):
                raise TypeError(
                    f'Invalid type "{type(value).__name__}"; expected '
                    f'"{cls.__name__}" or "dict".'
                )

            return cls(**value)

        return convert


class ListHookFactory:
    """
    A :class:`HookFactory` for :class:`list` and :class:`typing.List`.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is list or origin is list

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Iterable, Type[T]], T]:
        if not args:
            args = (Any,)
        item_type = args[0]

        def convert(value: Iterable, cls: Type[T]) -> T:
            value = converter.maybe_apply_strlist_hook(value)
            values = [converter.structure(v, item_type) for v in value]
            return list(values)  # type: ignore[return-value]

        return convert


class TupleHookFactory:
    """
    A :class:`HookFactory` for :class:`tuple` and :class:`typing.Tuple`.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is tuple or origin is tuple

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Union[Callable[[Iterable, Type[T]], T], Callable[[Sequence, Type[T]], T]]:
        if not args:
            args = (Any, ...)

        convert: Union[
            Callable[[Iterable, Type[T]], T],  # For list-like tuples
            Callable[[Sequence, Type[T]], T],  # For struct-like tuples
        ]
        if len(args) == 2 and args[1] == ...:
            item_type = args[0]

            def convert(value: Iterable, cls: Type[T]) -> T:
                value = converter.maybe_apply_strlist_hook(value)
                values = [converter.structure(v, item_type) for v in value]
                return tuple(values)  # type: ignore[return-value]

        else:

            def convert(value: Sequence, cls: Type[T]) -> T:
                value = converter.maybe_apply_strlist_hook(value)
                if len(value) != len(args):
                    raise TypeError(
                        f"Value must have {len(args)} items but has: {len(value)}"
                    )
                values = [converter.structure(v, t) for v, t in zip(value, args)]
                return tuple(values)  # type: ignore[return-value]

        return convert


class DictHookFactory:
    """
    A :class:`HookFactory` for :class:`dict` and :class:`typing.Dict`.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is dict or origin is dict

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Mapping, Type[T]], T]:
        if not args:
            args = (Any, Any)
        key_type, val_type = args

        def convert(value: Mapping, cls: Type[T]) -> T:
            values = {
                converter.structure(k, key_type): converter.structure(v, val_type)
                for k, v in value.items()
            }
            return values  # type: ignore[return-value]

        return convert


class MappingProxyTypeHookFactory:
    """
    A :class:`HookFactory` for :class:`types.MappingProxyType` (a read-only dict proxy).
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is MappingProxyType or origin is MappingProxyType

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Any, Type[T]], T]:
        if not args:
            args = (Any, Any)
        key_type, val_type = args

        def convert(value: Mapping, cls: Type[T]) -> T:
            values = {
                converter.structure(k, key_type): converter.structure(v, val_type)
                for k, v in value.items()
            }
            return MappingProxyType(values)  # type: ignore[return-value]

        return convert


class SetHookFactory:
    """
    A :class:`HookFactory` for :class:`set` and :class:`typing.Set`.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is set or origin is set

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Any, Type[T]], T]:
        if not args:
            args = (Any,)
        item_type = args[0]

        def convert(value: Iterable, cls: Type[T]) -> T:
            value = converter.maybe_apply_strlist_hook(value)
            values = [converter.structure(v, item_type) for v in value]
            return set(values)  # type: ignore[return-value]

        return convert


class FrozenSetHookFactory:
    """
    A :class:`HookFactory` for :class:`frozenset` and :class:`typing.FrozenSet`.
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return cls is frozenset or origin is frozenset

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Any, Type[T]], T]:
        if not args:
            args = (Any,)
        item_type = args[0]

        def convert(value: Iterable, cls: Type[T]) -> T:
            value = converter.maybe_apply_strlist_hook(value)
            values = [converter.structure(v, item_type) for v in value]
            return frozenset(values)  # type: ignore[return-value]

        return convert


class UnionHookFactory:
    """
    A :class:`HookFactory` for :data:`typing.Optional` and :data:`typing.Union`.

    If the input data already has one of the uniton types, it will be returned without
    further processing.  Otherwise, converters for all union types will be tried until
    one works (i.e., raises no exception).
    """

    @staticmethod
    def match(cls: type, origin: Optional[Any], args: Tuple[Any, ...]) -> bool:
        return origin in (Union, UnionType)

    @staticmethod
    def get_structure_hook(
        converter: TSConverter, cls: type, origin: Optional[Any], args: Tuple[Any, ...]
    ) -> Callable[[Any, Type[T]], T]:
        def convert(value: Any, cls: Type[T]) -> T:
            if type(value) in args:
                # Preserve val as-is if it already has a matching type.
                # Otherwise float(3.2) would be converted to int
                # if the converters are [int, float].
                return value
            for arg in args:
                try:
                    return converter.structure(value, arg)
                except Exception:  # noqa: S110
                    pass
            raise ValueError(f"Failed to convert value to any Union type: {value}")

        return convert



"""
Tests for "typed_settings.loaders".
"""

import dataclasses
import textwrap
from itertools import product
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import attrs
import pytest
from pytest import MonkeyPatch

from typed_settings.cls_utils import deep_options
from typed_settings.exceptions import (
    ConfigFileLoadError,
    ConfigFileNotFoundError,
    InvalidOptionsError,
    UnknownFormatError,
)
from typed_settings.loaders import (
    EnvLoader,
    FileFormat,
    FileLoader,
    InstanceLoader,
    OnePasswordLoader,
    PythonFormat,
    TomlFormat,
    clean_settings,
    tomllib,
)
from typed_settings.types import LoadedSettings, LoaderMeta, OptionList, SettingsDict

from .conftest import Host, Settings, SettingsClasses


class TestCleanSettings:
    """Tests for clean_settings."""

    def test_load_convert_dashes(self) -> None:
        """
        Dashes in settings and section names are replaced with underscores.
        """

        @dataclasses.dataclass(frozen=True)
        class Sub:
            b_1: str

        @dataclasses.dataclass(frozen=True)
        class Settings:
            a_1: str
            a_2: str
            sub_section: Sub

        s = {
            "a-1": "spam",
            "a_2": "eggs",
            "sub-section": {"b-1": "bacon"},
        }

        result = clean_settings(s, deep_options(Settings), "test")
        assert result == {
            "a_1": "spam",
            "a_2": "eggs",
            "sub_section": {"b_1": "bacon"},
        }

    def test_no_replace_dash_in_dict_keys(self) -> None:
        """
        "-" in TOML keys are replaced with "_" for sections and options, but
        "-" in actuall dict keys are left alone.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/3
        """

        @dataclasses.dataclass(frozen=True)
        class Settings:
            option_1: Dict[str, Any]
            option_2: Dict[str, Any]

        s = {
            "option-1": {"my-key": "val1"},
            "option-2": {"another-key": 23},
        }

        result = clean_settings(s, deep_options(Settings), "test")
        assert result == {
            "option_1": {"my-key": "val1"},
            "option_2": {"another-key": 23},
        }

    def test_invalid_settings(self) -> None:
        """
        Settings for which there is no attribute are errors.
        """
        s = {
            "url": "abc",
            "host": {"port": 23, "eggs": 42},
            "spam": 23,
        }
        with pytest.raises(InvalidOptionsError) as exc_info:
            clean_settings(s, deep_options(Settings), "t")
        assert str(exc_info.value) == ("Invalid options found in t: host.eggs, spam")

    def test_clean_settings_dict_values(self) -> None:
        """
        Some dicts may be actual values (not nested) classes.  Don't try to
        check theses as option paths.
        """

        @dataclasses.dataclass(frozen=True)
        class Settings:
            option: Dict[str, Any]

        s: SettingsDict = {"option": {"a": 1, "b": 2}}
        clean_settings(s, deep_options(Settings), "t")


class TestPythonFormat:
    """Tests for PythonFormat."""

    @pytest.mark.parametrize(
        "fmt, data",
        [
            (
                PythonFormat("example"),
                """\
                class example:
                    url = "spam"

                    class host:
                        port = 42
                """,
            ),
            (
                PythonFormat("EXAMPLE", key_transformer=PythonFormat.to_lower),
                """\
                class EXAMPLE:
                    URL = "spam"

                    class HOST:
                        PORT = 42
                """,
            ),
            (
                PythonFormat("example", flat=True),
                """\
                class example:
                    url = "spam"
                    host_port = 42
                """,
            ),
            (
                PythonFormat(
                    "EXAMPLE", key_transformer=PythonFormat.to_lower, flat=True
                ),
                """\
                class EXAMPLE:
                    URL = "spam"
                    HOST_PORT = 42
                """,
            ),
            (
                PythonFormat(None, key_transformer=PythonFormat.to_lower),
                """\
                URL = "spam"

                class HOST:
                    PORT = 42
                """,
            ),
        ],
    )
    def test_load_python(self, fmt: FileFormat, data: str, tmp_path: Path) -> None:
        """
        We can load settings from a Python file.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text(textwrap.dedent(data))
        result = fmt(config_file, Settings, deep_options(Settings))
        assert result == {
            "url": "spam",
            "host": {"port": 42},
        }

    @pytest.mark.parametrize("section", ["example", "spam.example"])
    def test_section_not_found(self, section: str, tmp_path: Path) -> None:
        """
        An empty dict is returned when the config file does not contain the
        desired class.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text("class spam:\n    a = 'spam'\n")
        result = PythonFormat(section)(config_file, Settings, deep_options(Settings))
        assert result == {}

    def test_file_not_found(self) -> None:
        """
        "ConfigFileNotFoundError" is raised when a file does not exist.
        """
        pytest.raises(
            ConfigFileNotFoundError,
            PythonFormat(""),
            Path("x"),
            deep_options(Settings),
            Settings,
        )

    def test_file_invalid(self, tmp_path: Path) -> None:
        """
        "ConfigFileLoadError" is raised when a file contains invalid Python.
        """
        config_file = tmp_path.joinpath("settings.py")
        config_file.write_text("3x = 'spam")
        pytest.raises(
            ConfigFileLoadError,
            PythonFormat(""),
            config_file,
            deep_options(Settings),
            Settings,
        )


class TestTomlFormat:
    """Tests for TomlFormat."""

    @pytest.mark.parametrize(
        "fmt, data",
        [
            (
                TomlFormat("example"),
                """\
                [example]
                url = "spam"
                [example.host]
                port = 42
                """,
            ),
        ],
    )
    def test_load_toml(self, fmt: FileFormat, data: str, tmp_path: Path) -> None:
        """
        We can load settings from a TOML file.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(textwrap.dedent(data))
        result = fmt(config_file, Settings, deep_options(Settings))
        assert result == {
            "url": "spam",
            "host": {"port": 42},
        }

    def test_load_from_nested(self, tmp_path: Path) -> None:
        """
        We can load settings from a nested section (e.g., "tool.example").
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool.example]
            a = "spam"
            [tool.example.sub]
            b = "eggs"
        """
        )
        result = TomlFormat("tool.example")(
            config_file,
            Settings,
            deep_options(Settings),
        )
        assert result == {
            "a": "spam",
            "sub": {"b": "eggs"},
        }

    def test_load_no_section(self, tmp_path: Path) -> None:
        """
        If the sections is ``None``, the "top level" settings are loaded.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """\
            a = "spam"
            [example]
            b = "eggs"
        """
        )
        result = TomlFormat(None)(
            config_file,
            Settings,
            deep_options(Settings),
        )
        assert result == {
            "a": "spam",
            "example": {
                "b": "eggs",
            },
        }

    @pytest.mark.parametrize("section", ["example", "tool.example"])
    def test_section_not_found(self, section: str, tmp_path: Path) -> None:
        """
        An empty dict is returned when the config file does not contain the
        desired section.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool]
            a = "spam"
        """
        )
        result = TomlFormat(section)(config_file, Settings, deep_options(Settings))
        assert result == {}

    def test_file_not_found(self) -> None:
        """
        "ConfigFileNotFoundError" is raised when a file does not exist.
        """
        pytest.raises(
            ConfigFileNotFoundError,
            TomlFormat(""),
            Path("x"),
            deep_options(Settings),
            Settings,
        )

    def test_file_not_allowed(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        "ConfigFileLoadError" is raised when a file cannot be accessed.
        """

        def toml_load(path: Path) -> None:
            raise PermissionError()

        monkeypatch.setattr(tomllib, "load", toml_load)

        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[tool]
            a = "spam"
        """
        )

        pytest.raises(
            ConfigFileLoadError,
            TomlFormat(""),
            config_file,
            deep_options(Settings),
            Settings,
        )

    def test_file_invalid(self, tmp_path: Path) -> None:
        """
        "ConfigFileLoadError" is raised when a file contains invalid TOML.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text("spam")
        pytest.raises(
            ConfigFileLoadError,
            TomlFormat(""),
            config_file,
            deep_options(Settings),
            Settings,
        )


class TestFileLoader:
    """Tests for FileLoader."""

    @pytest.fixture
    def fnames(self, tmp_path: Path) -> List[Path]:
        p0 = tmp_path.joinpath("0.toml")
        p1 = tmp_path.joinpath("1.toml")
        p2 = tmp_path.joinpath("2")
        p3 = tmp_path.joinpath("3")
        p0.touch()
        p2.touch()
        return [p0, p1, p2, p3]

    @pytest.mark.parametrize(
        "cfn, env, expected",
        [
            ([], None, []),
            ([0], None, [0]),
            ([1], None, []),
            ([2], None, [2]),
            ([3], None, []),
            ([], [0], [0]),
            ([0, 1], [2, 3], [0, 2]),
            ([2, 1, 0], [2], [2, 0, 2]),
        ],
    )
    def test_get_config_filenames(
        self,
        cfn: List[int],
        env: Optional[List[int]],
        expected: List[int],
        fnames: List[Path],
        monkeypatch: MonkeyPatch,
    ) -> None:
        """
        Config files names (cfn) can be specified explicitly or via an env var.
        It's no problem if a files does not exist.
        """
        var: Optional[str]
        if env is not None:
            monkeypatch.setenv("CF", ":".join(str(fnames[i]) for i in env))
            var = "CF"
        else:
            var = None

        paths = FileLoader._get_config_filenames([fnames[i] for i in cfn], var)
        assert paths == [fnames[i] for i in expected]

    def test_get_config_filenames_empty_fn(
        self,
        fnames: List[Path],
        monkeypatch: MonkeyPatch,
    ) -> None:
        """
        Empty filenames from the env var are ignored.
        """
        monkeypatch.setenv("CF", f"::{fnames[0]}:")
        paths = FileLoader._get_config_filenames([], "CF")
        assert paths == fnames[:1]

    def test_load_file(self, tmp_path: Path) -> None:
        """
        Settings are cleaned for each file individually.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            le-option = "spam"
        """
        )

        @dataclasses.dataclass(frozen=True)
        class Settings:
            le_option: str = ""

        loader = FileLoader(
            formats={"*.toml": TomlFormat("le-section")},
            files=[config_file],
        )
        s = loader._load_file(config_file, Settings, deep_options(Settings))
        assert s == {"le_option": "spam"}

    def test_load_file2(self, tmp_path: Path) -> None:
        """
        Settings are cleaned for each file individually.  In that process,
        "-" is normalized to "_".  This may result in duplicate settings and
        the last one wins in that case.
        """
        config_file = tmp_path.joinpath("settings.toml")
        config_file.write_text(
            """[le-section]
            le_option = "eggs"
            le-option = "spam"
        """
        )

        @dataclasses.dataclass(frozen=True)
        class Settings:
            le_option: str = ""

        loader = FileLoader(
            formats={"*.toml": TomlFormat("le-section")},
            files=[config_file],
        )
        s = loader._load_file(config_file, Settings, deep_options(Settings))
        assert s == {"le_option": "spam"}

    def test_load_file_invalid_format(self) -> None:
        """
        An error is raised if a file has an unknown extension.
        """
        loader = FileLoader({"*.toml": TomlFormat("t")}, [])
        pytest.raises(UnknownFormatError, loader._load_file, Path("f.py"), [], type)

    def test_load(self, tmp_path: Path) -> None:
        """
        FileLoader() loads multiple files, each one overriding options
        from its predecessor.
        """
        cf1 = tmp_path.joinpath("s1.toml")
        cf1.write_text(
            """[le-section]
            le-spam = "spam"
            le-eggs = "spam"
        """
        )
        cf2 = tmp_path.joinpath("s2.toml")
        cf2.write_text(
            """[le-section]
            le_eggs = "eggs"
        """
        )

        @dataclasses.dataclass(frozen=True)
        class Settings:
            le_spam: str = ""
            le_eggs: str = ""

        loader = FileLoader({"*.toml": TomlFormat("le-section")}, [cf1, cf2])
        s = loader(Settings, deep_options(Settings))
        assert s == [
            LoadedSettings(
                {"le_spam": "spam", "le_eggs": "spam"},
                LoaderMeta(f"FileLoader[{cf1}]", base_dir=cf1.parent),
            ),
            LoadedSettings(
                {"le_eggs": "eggs"},
                LoaderMeta(f"FileLoader[{cf2}]", base_dir=cf2.parent),
            ),
        ]

    @pytest.mark.parametrize(
        "is_mandatory, is_path, in_env, exists",
        product([True, False], repeat=4),
    )
    def test_mandatory_files(
        self,
        is_mandatory: bool,
        is_path: bool,
        in_env: bool,
        exists: bool,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """
        Paths with a "!" are mandatory and an error is raised if they don't
        exist.
        """
        path = tmp_path.joinpath("s.toml")
        if exists:
            path.touch()
        p: Union[Path, str] = f"!{path}" if is_mandatory else str(path)
        if is_path:
            p = Path(p)
        files = []
        if in_env:
            monkeypatch.setenv("TEST_SETTINGS", str(p))
        else:
            files = [p]

        loader = FileLoader({"*": TomlFormat("test")}, files, "TEST_SETTINGS")
        if is_mandatory and not exists:
            pytest.raises(FileNotFoundError, loader, Settings, [])
        else:
            loader(Settings, ())


class TestEnvLoader:
    """Tests for EnvLoader."""

    def test_from_env(self, monkeypatch: MonkeyPatch) -> None:
        """
        Load options from env vars, ignore env vars for which no settings
        exist.
        """
        monkeypatch.setenv("T_URL", "foo")
        monkeypatch.setenv("T_HOST", "spam")  # Haha! Just a deceit!
        monkeypatch.setenv("T_HOST_PORT", "25")
        loader = EnvLoader(prefix="T_")
        results = loader(Settings, deep_options(Settings))
        assert results == LoadedSettings(
            {"url": "foo", "host": {"port": "25"}},
            LoaderMeta("EnvLoader"),
        )

    def test_no_env_prefix(self, monkeypatch: MonkeyPatch) -> None:
        """
        It is okay to use an empty prefix.
        """
        monkeypatch.setenv("URL", "spam")

        loader = EnvLoader(prefix="")
        results = loader(Settings, deep_options(Settings))
        assert results == LoadedSettings({"url": "spam"}, LoaderMeta("EnvLoader"))

    @pytest.mark.parametrize("delimiter", ["_", "__", "#"])
    def test_nested_delimiter(self, delimiter: str, monkeypatch: MonkeyPatch) -> None:
        """
        The delimiter for for name parts of nested settings can be set by the user.
        """
        monkeypatch.setenv(f"T_HOST{delimiter}NAME", "test")
        loader = EnvLoader(prefix="T_", nested_delimiter=delimiter)
        results = loader(Settings, deep_options(Settings))
        assert results == LoadedSettings(
            {"host": {"name": "test"}}, LoaderMeta("EnvLoader")
        )


class TestInstanceLoader:
    """Tests for InstanceLoader."""

    def test_from_inst(
        self, settings_clss: SettingsClasses, options: OptionList
    ) -> None:
        """
        Load options from env vars, ignore env vars for which no settings
        exist.
        """
        Settings, Host = settings_clss
        inst = Settings(host=Host(name="spam", port=42), url="eggs", default=23)
        loader = InstanceLoader(inst)
        results = loader(Settings, deep_options(Settings))
        assert results == LoadedSettings(
            {
                "default": 23,
                "url": "eggs",
                "host": {"name": "spam", "port": 42},
            },
            LoaderMeta("InstanceLoader"),
        )

    def test_invalid_type(self) -> None:
        """
        A ValueError is raised if the "instance" object is not an instances of the
        correct settings class.
        """

        @attrs.define
        class SettingsNoDc:
            host: Host
            url: str
            default: int

        inst = SettingsNoDc(Host("spam", 42), "eggs", 23)
        loader = InstanceLoader(inst)
        pytest.raises(ValueError, loader, Settings, deep_options(Settings))


class TestOnePasswordLoader:
    """Tests for OnePasswordLoader."""

    def test_load(self, mock_op: None) -> None:
        """
        Settings can be loaded from 1Password.
        """

        @dataclasses.dataclass(frozen=True)
        class Settings:
            username: str
            password: str
            is_admin: bool = False

        loader = OnePasswordLoader(item="Test", vault="Test")
        s = loader(Settings, deep_options(Settings))
        assert s == LoadedSettings(
            {"username": "spam", "password": "eggs"}, LoaderMeta("OnePasswordLoader")
        )



"""
Utilities for generating Click options.
"""

from datetime import date, datetime, timedelta
from enum import Enum
from functools import partial, update_wrapper
from pathlib import Path
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import click
from click.core import ParameterSource

from . import _core, cls_utils, converters
from .cli_utils import (
    NO_DEFAULT,
    Default,
    DefaultFactorySentinel,
    StrDict,
    TypeArgsMaker,
    TypeHandlerFunc,
    get_default,
)
from .constants import CLICK_METADATA_KEY as METADATA_KEY
from .converters import Converter, default_converter
from .loaders import EnvLoader, Loader
from .processors import Processor
from .types import (
    SECRET_REPR,
    ST,
    LoadedValue,
    LoaderMeta,
    MergedSettings,
    OptionInfo,
    OptionList,
    SecretStr,
    SettingsClass,
)


__all__ = [
    "click_options",
    "pass_settings",
    "DecoratorFactory",
    "ClickOptionFactory",
    "OptionGroupFactory",
    "handle_datetime",
    "handle_enum",
    "DEFAULT_TYPES",
    "ClickHandler",
    "F",
]


CTX_KEY = "settings"


Callback = Callable[[click.Context, click.Option, Any], Any]
F = TypeVar("F", bound=Callable[..., Any])
"""TypeVar for arbitrary functions."""
Decorator = Callable[[F], F]


def click_options(
    settings_cls: Type[ST],
    loaders: Union[str, Sequence[Loader]],
    *,
    processors: Sequence[Processor] = (),
    converter: Optional[Converter] = None,
    base_dir: Path = Path(),
    type_args_maker: Optional[TypeArgsMaker] = None,
    argname: Optional[str] = None,
    decorator_factory: "Optional[DecoratorFactory]" = None,
    show_envvars_in_help: bool = False,
    reload_settings_on_invoke: bool = False,
) -> Callable[[F], F]:
    r"""
    **Decorator:** Generate :mod:`click` options for a CLI which override
    settings loaded via :func:`.load_settings()`.

    Args:
        settings_cls: The settings class to generate options for.

        loaders: Either a string with your app name or a list of
            :class:`.Loader`\ s.  If it's a string, use it with
            :func:`~typed_settings.default_loaders()` to get the default
            loaders.

        processors: A list of settings :class:`.Processor`'s.

        converter: An optional :class:`.Converter` used for converting
            option values to the required type.

            By default, :data:`typed_settings.default_converter()` is used.

        base_dir: Base directory for resolving relative paths in default option values.

        type_args_maker: The type args maker that is used to generate keyword
            arguments for :func:`click.option()`.  By default, use
            :class:`.TypeArgsMaker` with :class:`ClickHandler`.

        argname: An optional argument name for the settings instance that is
            passed to the CLI.  If it is set, the settings instances is no
            longer passed as positional argument but as key word argument.

            This allows a CLI function to be decorated with this function
            multiple times.

        decorator_factory: A class that generates Click decorators for options
            and settings classes.  This allows you to, e.g., use
            `option groups`_ via :class:`OptionGroupFactory`.  The default
            generates normal Click options via :class:`ClickOptionFactory`.

            .. _option groups: https://click-option-group.readthedocs.io

        show_envvars_in_help: If ``True`` and if the
            :class:`~typed_settings.loaders.EnvLoader` is being used, show the
            names of the environment variable a value is loaded from.

        reload_settings_on_invoke: By default, the default values will be loaded (from
            config files and env vars) when the CLI is created.  If you set this to
            ``True``, the defaults are reloaded when the CLI is invoked.  This makes
            running the CLI slower but can improve testability, because you can change
            the values of env vars in tests without needing to create a new CLI
            function.

    Return:
        A decorator for a click command.

    Raise:
        InvalidSettingsError: If an instance of *cls* cannot be created for the given
            settings.

    Example:
        .. code-block:: python

           import click
           import typed_settings as ts

           @ts.settings
           class Settings: ...

           @click.command()
           @ts.click_options(Settings, "example")
           def cli(settings: Settings) -> None:
               print(settings)

    .. versionchanged:: 1.0.0
       Instead of a list of loaders, you can also just pass an application
       name.
    .. versionchanged:: 1.1.0
       Added the *argname* parameter.
    .. versionchanged:: 1.1.0
       Added the *decorator_factory* parameter.
    .. versionchanged:: 2.0.0
       Renamed *type_handler* to *type_args_maker* and changed it's type to
       ``TypeArgsMaker``.
    .. versionchanged:: 23.0.0
       Made *converter*, *type_args_maker*, *argname*, and *decorator_factory*
       a keyword-only argument
    .. versionchanged:: 23.0.0
       Added the *processors* argument
    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    .. versionchanged:: 24.5.0
       Added the *reload_settings_on_invoke* argument
    """
    if isinstance(loaders, str):
        loaders = _core.default_loaders(loaders)

    env_loader: Optional[EnvLoader] = None
    if show_envvars_in_help:
        _loaders = [ldr for ldr in loaders if isinstance(ldr, EnvLoader)]
        if _loaders:
            env_loader = _loaders[-1]

    converter = converter or default_converter()
    state = _core.SettingsState(settings_cls, loaders, processors, converter, base_dir)
    grouped_options = cls_utils.group_options(state.settings_class, state.options)
    type_args_maker = type_args_maker or TypeArgsMaker(ClickHandler())
    decorator_factory = decorator_factory or ClickOptionFactory()

    ts_decorator = _get_ts_decorator(
        state,
        grouped_options,
        type_args_maker,
        argname,
        decorator_factory,
        env_loader,
        reload_settings_on_invoke=reload_settings_on_invoke,
    )
    return ts_decorator


def _get_ts_decorator(  # noqa: C901
    state: _core.SettingsState[ST],
    grouped_options: List[Tuple[type, OptionList]],
    type_args_maker: TypeArgsMaker,
    argname: Optional[str],
    decorator_factory: "DecoratorFactory",
    env_loader: Optional[EnvLoader],
    reload_settings_on_invoke: bool,
) -> Callable[[F], F]:
    """
    Return a wrapper function that decorates the wrapped function with the Click
    decorators.
    """

    def ts_decorator(orig_f: F) -> F:
        """
        The wrapper that actually decorates a function with all options.
        """
        # If "reload_settings_on_invoke" is True, these initial default settings will
        # be reloaded on each invocation.  The help texts remain unchanged, though.
        initial_default_settings = _core._load_settings(state)

        # Create a *cls* instances from the settings dict stored in
        # :attr:`click.Context.obj` and passes it to the decorated function *orig_f*.
        # def new_func(*args: "P.args", **kwargs: "P.kwargs") -> "R":
        def new_func(*args: Any, **kwargs: Any) -> Any:
            if reload_settings_on_invoke:
                default_settings = _core._load_settings(state)
            else:
                default_settings = initial_default_settings
            ctx = click.get_current_context()
            if ctx.obj is None:
                ctx.obj = {}
            meta = LoaderMeta("Command line args")
            cli_options = ctx.obj.get(CTX_KEY, {})
            cli_settings: MergedSettings = {}
            for option in state.options:
                path = option.path
                if path in cli_options:  # pragma: no cover
                    # "path" *should* always be in "cli_options", b/c we *currently*
                    # generate CLI options for all options.  But let's stay safe here
                    # in case the behavior changes in the future.
                    cli_settings[path] = LoadedValue(cli_options[path], meta)
                elif path in default_settings:
                    cli_settings[path] = default_settings[path]
                # else:
                #     This is the case, when the default should be used and it is a
                #     factory function.  The _DefaultsLoader did not invoke it yet.
            settings = _core.convert(cli_settings, state)
            if argname:
                ctx_key = argname
                kwargs = {argname: settings, **kwargs}  # type: ignore
            else:
                ctx_key = CTX_KEY
                args = (settings, *args)  # type: ignore
            ctx.obj[ctx_key] = settings
            return orig_f(*args, **kwargs)

        wrapped_f = cast(F, update_wrapper(new_func, orig_f))

        option_decorator = decorator_factory.get_option_decorator()
        for g_cls, g_opts in reversed(grouped_options):
            for oinfo in reversed(g_opts):
                default = get_default(oinfo, initial_default_settings, state.converter)
                envvar = env_loader.get_envvar(oinfo) if env_loader else None
                option = _mk_option(
                    option_decorator,  # type: ignore[arg-type]
                    oinfo,
                    default,
                    type_args_maker,
                    envvar,
                )
                wrapped_f = option(wrapped_f)  # type: ignore[assignment,arg-type]
            wrapped_f = decorator_factory.get_group_decorator(g_cls)(
                wrapped_f  # type: ignore[arg-type]
            )

        return wrapped_f

    return ts_decorator


@overload
def pass_settings(
    f: None = None, *, argname: Optional[str] = ...
) -> Callable[[F], F]: ...


@overload
def pass_settings(f: F, *, argname: Optional[str] = ...) -> F: ...


def pass_settings(
    f: Optional[F] = None,
    *,
    argname: Optional[str] = None,
) -> Union[F, Callable[[F], F]]:
    """
    **Decorator:** Mark a callback as wanting to receive the innermost settings
    instance as first argument.

    If you specify an *argname* in :func:`click_options()`, you must specify
    the same name here in order to get the correct settings instance.  The
    settings instance is then passed as keyword argument.

    Args:
        f: If this decorator is applied without any arguments, this is the to be
            decrorated function.  If you pass any keyword arguments, this is ``None``.
        argname: An optional argument name.  If it is set, the settings
            instance is no longer passed as positional argument but as key
            word argument.

    Return:
        A decorator for a click command.

    Example:
        .. code-block:: python

           import click
           import typed_settings as ts

           @ts.settings
           class Settings: ...

           @click.group()
           @click_options(Settings, "example", argname="my_settings")
           def cli(my_settings):
               pass

           @cli.command()
           # Use the same "argname" as above!
           @pass_settings(argname="my_settings")
           def sub_cmd(*, my_settings):
               print(my_settings)

    .. versionchanged:: 1.1.0
       Add the *argname* parameter.
    """
    ctx_key = argname or CTX_KEY

    def decorator(f: F) -> F:
        def new_func(*args: Any, **kwargs: Any) -> Any:
            ctx = click.get_current_context()
            node: Optional[click.Context] = ctx
            settings = None
            while node is not None:
                if isinstance(node.obj, dict) and ctx_key in node.obj:
                    settings = node.obj[ctx_key]
                    break
                node = node.parent

            if argname:
                kwargs = {argname: settings, **kwargs}
            else:
                args = (settings, *args)

            return ctx.invoke(f, *args, **kwargs)

        return cast(F, update_wrapper(new_func, f))

    if f is None:
        return decorator

    return decorator(f)


class TSOption(click.Option):
    def value_from_envvar(self, ctx: click.Context) -> Optional[Any]:
        return None


class DecoratorFactory(Protocol):
    """
    **Protocol:** Methods that a Click decorator factory must implement.

    The decorators returned by the procol methods are used to construct the
    Click options and possibly option groups.

    .. versionadded:: 1.1.0
    """

    def get_option_decorator(self) -> Callable[..., Decorator[F]]:
        """
        Return the decorator that is used for creating Click options.

        It must be compatible with :func:`click.option()`.
        """
        ...

    def get_group_decorator(self, settings_cls: type) -> Decorator[F]:
        """
        Return a decorator for the current settings class.

        This can, e.g., be used to group option by settings class.
        """
        ...


class ClickOptionFactory:
    """
    Factory for default Click decorators.
    """

    def get_option_decorator(self) -> Callable[..., Decorator[F]]:
        """
        Return :func:`click.option()`.
        """
        return partial(click.option, cls=TSOption)

    def get_group_decorator(self, settings_cls: SettingsClass) -> Decorator[F]:
        """
        Return a no-op decorator that leaves the decorated function unchanged.
        """
        return lambda f: f


class OptionGroupFactory:
    """
    Factory got generating Click option groups via
    https://click-option-group.readthedocs.io.
    """

    def __init__(self) -> None:
        try:
            from click_option_group import GroupedOption, optgroup
        except ImportError as e:
            raise ModuleNotFoundError(
                "Module 'click_option_group' not installed.  Please run "
                "'python -m pip install -U typed-settings[option-groups]'"
            ) from e
        self.optgroup = optgroup

        class TSGroupedOption(GroupedOption):
            def value_from_envvar(self, ctx: click.Context) -> Optional[Any]:
                return None

        self.opt_cls = TSGroupedOption

    def get_option_decorator(self) -> Callable[..., Decorator[F]]:
        """
        Return :class:`click_option_group.optgroup` option.
        """
        return partial(self.optgroup.option, cls=self.opt_cls)

    def get_group_decorator(self, settings_cls: SettingsClass) -> Decorator[F]:
        """
        Return a :class:`click_option_group.optgroup` group instantiated with
        the first line of *settings_cls*'s docstring.
        """
        try:
            name = settings_cls.__doc__.strip().splitlines()[0]  # type: ignore
        except (AttributeError, IndexError):
            name = f"{settings_cls.__name__} options"
        return cast(Decorator[F], self.optgroup.group(name))


def handle_datetime(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Use :class:`click.DateTime` as option type and convert the default value
    to an ISO string.
    """
    kwargs: StrDict = {
        "type": click.DateTime(
            ["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S%z"]
        ),
    }
    if isinstance(default, datetime):
        kwargs["default"] = default.isoformat()
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_date(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Use :class:`click.DateTime` as option type and convert the default value
    to an ISO string.
    """
    typ = partial(converters.to_date, cls=date)
    typ.__name__ = converters.to_date.__name__  # type: ignore[attr-defined]
    kwargs: StrDict = {
        "type": typ,
        "metavar": "[%Y-%m-%d]",
    }
    if isinstance(default, date):
        kwargs["default"] = default.isoformat()
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_timedelta(type: type, default: Default, is_optional: bool) -> StrDict:
    """
    Use :class:`click.DateTime` as option type and convert the default value
    to an ISO string.
    """
    typ = partial(converters.to_timedelta, cls=timedelta)
    typ.__name__ = converters.to_timedelta.__name__  # type: ignore[attr-defined]
    kwargs: StrDict = {
        "type": typ,
        "metavar": "[-][Dd][HHh][MMm][SS[.ffffff]s]",
    }
    if isinstance(default, timedelta):
        kwargs["default"] = converters.timedelta_to_str(default)
    elif is_optional:
        kwargs["default"] = None
    return kwargs


def handle_enum(type: Type[Enum], default: Default, is_optional: bool) -> StrDict:
    """
    Use :class:`click.Choice` as option type and use the enum value's name as
    default.
    """
    kwargs: StrDict = {"type": click.Choice(list(type.__members__))}
    if isinstance(default, type):
        # Convert Enum instance to string
        kwargs["default"] = default.name
    elif is_optional:
        kwargs["default"] = None

    return kwargs


#: Default handlers for click option types.
DEFAULT_TYPES: Dict[type, TypeHandlerFunc] = {
    datetime: handle_datetime,
    date: handle_date,
    timedelta: handle_timedelta,
    Enum: handle_enum,
}


class ClickHandler:
    """
    Implementation of the :class:`~typed_settings.cli_utils.TypeHandler`
    protocol for Click.

    Args:
        extra_types: A dict mapping types to specialized handler functions.
            Use :data:`DEFAULT_TYPES` by default.

    .. versionadded:: 2.0.0
    """

    def __init__(
        self, extra_types: Optional[Dict[type, TypeHandlerFunc]] = None
    ) -> None:
        self.extra_types = extra_types or DEFAULT_TYPES

    def get_scalar_handlers(self) -> Dict[type, TypeHandlerFunc]:
        return self.extra_types

    def handle_scalar(
        self,
        type: Optional[type],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        kwargs: StrDict = {"type": type}
        if default not in (None, NO_DEFAULT):
            kwargs["default"] = default
        elif is_optional:
            kwargs["default"] = None
        if type:
            if issubclass(type, bool):
                kwargs["is_flag"] = True
            elif issubclass(type, SecretStr):
                kwargs["metavar"] = "TEXT"

        return kwargs

    def handle_tuple(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Optional[Tuple],
        is_optional: bool,
    ) -> StrDict:
        kwargs = {
            "type": types,
            "nargs": len(types),
            "default": default,
        }
        return kwargs

    def handle_collection(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Optional[Collection[Any]],
        is_optional: bool,
    ) -> StrDict:
        kwargs = type_args_maker.get_kwargs(types[0], NO_DEFAULT)
        kwargs["default"] = default
        kwargs["multiple"] = True
        return kwargs

    def handle_mapping(
        self,
        type_args_maker: TypeArgsMaker,
        types: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        def cb(
            ctx: click.Context,
            param: click.Option,
            value: Optional[Iterable[str]],
        ) -> Optional[Dict[str, str]]:
            if not value:
                return {}
            splitted = [v.partition("=") for v in value]
            items = {k: v for k, _, v in splitted}
            return items

        kwargs = {
            "metavar": "KEY=VALUE",
            "multiple": True,
            "callback": cb,
        }
        if not isinstance(default, Mapping):
            default = {}
        kwargs["default"] = [f"{k}={v}" for k, v in default.items()]

        return kwargs


def _mk_option(
    option_fn: Callable[..., Decorator[F]],
    oinfo: OptionInfo,
    default: Default,
    type_args_maker: TypeArgsMaker,
    envvar: Optional[str],
) -> Decorator[F]:
    """
    Recursively creates click options and returns them as a list.
    """
    user_config = dict(oinfo.metadata.get(METADATA_KEY, {}))

    # The option type specifies the default option kwargs
    kwargs = type_args_maker.get_kwargs(oinfo.cls, default)
    if envvar:
        kwargs["envvar"] = envvar
        kwargs["show_envvar"] = True

    param_decls: Tuple[str, ...]
    user_param_decls: Union[str, Sequence[str]]
    user_param_decls = user_config.pop("param_decls", ())
    if not user_param_decls:
        option_name = oinfo.path.replace(".", "-").replace("_", "-")
        if kwargs.get("is_flag"):
            param_decls = (f"--{option_name}/--no-{option_name}",)
        else:
            param_decls = (f"--{option_name}",)
    elif isinstance(user_param_decls, str):
        param_decls = (user_param_decls,)
    else:
        param_decls = tuple(user_param_decls)

    # The type's kwargs should not be able to set these values since they are
    # needed for everything to work:
    kwargs["show_default"] = True
    kwargs["expose_value"] = False
    kwargs["callback"] = _make_callback(
        oinfo.path, kwargs.get("callback"), user_config.pop("callback", None)
    )

    # Get "help" from the user_config *now*, because we may need to update it
    # below.  Also replace "None" with "".
    kwargs["help"] = user_config.pop("help", None) or ""

    if "default" in kwargs:  # pragma: no cover
        if oinfo.is_secret:
            kwargs["show_default"] = SECRET_REPR
    else:
        kwargs["required"] = True

    # The user has the last word, though.
    kwargs.update(user_config)

    return option_fn(*param_decls, **kwargs)


def _make_callback(
    path: str,
    type_callback: Optional[Callback],
    user_callback: Optional[Callback],
) -> Callback:
    """
    Generate a callback that adds option values to the settings instance in the
    context.

    It also calls a type's callback if there should be one.
    """

    def cb(ctx: click.Context, param: click.Option, value: Any) -> Any:
        param_source = ctx.get_parameter_source(param.name or "")

        # We may not want to add "value" to the settings dict for various reasons:
        param_source_default = param_source is ParameterSource.DEFAULT
        if param_source_default and isinstance(param.default, DefaultFactorySentinel):
            # Don't add the value to the settings dict if it has a default factory.
            # The factory should be invoked later.
            return value
        if param_source_default and value not in (None, (), {}):
            # Don't add default values (that come from loaded settings) because this
            # would override the original LoaderMeta.
            # Only make an exception for "None" and empty containers to handle
            # "Optional[T]" with no explicit default (e.g., "myflag: bool | None").
            return value

        if type_callback is not None:
            value = type_callback(ctx, param, value)
        if user_callback is not None:
            value = user_callback(ctx, param, value)

        if ctx.obj is None:
            ctx.obj = {}
        settings = ctx.obj.setdefault(CTX_KEY, {})
        settings[path] = value

        return value

    return cb



"""
Tests for "typed_settings.argparse_utils".
"""

import sys
from pathlib import Path
from typing import Any, Callable, List, TypeVar

import attrs
import pytest

from typed_settings import (
    _compat,
    cli_argparse,
    cli_utils,
    constants,
    default_converter,
    default_loaders,
    option,
    settings,
)


T = TypeVar("T")


Invoke = Callable[..., Any]


Cli = Callable[[], Any]


@settings
class Settings:
    """A simple settings class for testing."""

    o: int


@pytest.fixture(name="invoke")
def _invoke(monkeypatch: pytest.MonkeyPatch) -> Invoke:
    def invoke(cli: Callable[[], Any], *args: str) -> Any:
        with monkeypatch.context() as m:
            m.setattr(sys, "argv", [cli.__name__, *list(args)])
            return cli()

    return invoke


def test_cli(invoke: Invoke) -> None:
    """
    Basic test "cli()" - simple CLI for a simple settings class.
    """

    @cli_argparse.cli(Settings, "test")
    def cli(settings: Settings) -> None:
        assert settings == Settings(3)

    invoke(cli, "--o=3")


def test_cli_explicit_config(invoke: Invoke) -> None:
    """
    Basic test "cli()" with explicit loaders, converter config.
    """
    loaders = default_loaders("test")
    converter = default_converter()
    tam = cli_utils.TypeArgsMaker(cli_argparse.ArgparseHandler())

    @cli_argparse.cli(
        Settings,
        loaders=loaders,
        converter=converter,
        type_args_maker=tam,
    )
    def cli(settings: Settings) -> None:
        assert settings == Settings(3)

    invoke(cli, "--o=3")


def test_cli_desc_from_func(invoke: Invoke, capsys: pytest.CaptureFixture) -> None:
    """
    The CLI function's docstring is used as argparse CLI description.
    """

    @cli_argparse.cli(Settings, "test")
    def cli(settings: Settings) -> None:
        """
        Le description.
        """

    with pytest.raises(SystemExit):
        invoke(cli, "--help")

    out, err = capsys.readouterr()
    assert out.startswith("usage: cli [-h] --o INT\n\nLe description.\n")
    assert err == ""


def test_cli_desc_from_kwarg(invoke: Invoke, capsys: pytest.CaptureFixture) -> None:
    """
    The argparse CLI description can be explicitly configured with a keyword arg.
    """

    @cli_argparse.cli(Settings, "test", description="Le description")
    def cli(settings: Settings) -> None:
        """
        spam spam spam.
        """

    with pytest.raises(SystemExit):
        invoke(cli, "--help")

    out, err = capsys.readouterr()
    assert out.startswith("usage: cli [-h] --o INT\n\nLe description\n")
    assert err == ""


def test_manual_parser() -> None:
    """
    Basic test for "make_parser()" and "namespace2settings"().
    """
    parser, merged_settings = cli_argparse.make_parser(Settings, "test")
    namespace = parser.parse_args(["--o", "3"])
    result = cli_argparse.namespace2settings(
        Settings, namespace, merged_settings=merged_settings
    )
    assert result == Settings(3)


def test_manual_parser_explicit_config() -> None:
    """
    Basic test for "make_parser()" and "namespace2settings"() with explicit
    config.
    """
    loaders = default_loaders("test")
    converter = default_converter()
    tam = cli_utils.TypeArgsMaker(cli_argparse.ArgparseHandler())
    parser, merged_settings = cli_argparse.make_parser(
        Settings,
        loaders=loaders,
        converter=converter,
        type_args_maker=tam,
    )
    namespace = parser.parse_args(["--o", "3"])
    result = cli_argparse.namespace2settings(
        Settings,
        namespace,
        merged_settings=merged_settings,
        converter=converter,
    )
    assert result == Settings(3)


def test_invalid_bool_flag() -> None:
    """
    Only "long" boolean flags (--flag) are supported, but not short ones (-f).
    """

    @settings
    class Settings:
        flag: bool = option(argparse={"param_decls": ("-f")})

    with pytest.raises(ValueError, match="boolean flags.*--.*supported"):
        cli_argparse.make_parser(Settings, "test")


def test_attrs_meta_not_modified() -> None:
    """
    The attrs meta data with with user defined argparse config is not modified.

    Regression test for #29.
    """

    @settings
    class S:
        opt: int = option(help="spam", argparse={"param_decls": "-o"})

    meta = attrs.fields(S).opt.metadata[constants.METADATA_KEY]

    assert meta[cli_argparse.METADATA_KEY] == {"help": "spam", "param_decls": "-o"}

    cli_argparse.make_parser(S, "test")
    cli_argparse.make_parser(S, "test")

    assert meta[cli_argparse.METADATA_KEY] == {"help": "spam", "param_decls": "-o"}


def test_resolve_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, invoke: Invoke
) -> None:
    """
    Relative paths passed via the command line are resolved based on the user's CWD.
    """

    @settings
    class Settings:
        a: Path = Path("default")
        b: Path = Path("default")  # Load from file
        c: Path = Path("default")  # Load from env var
        d: Path = Path("default")  # Load from cli arg

    spath = tmp_path.joinpath("settings.toml")
    spath.write_text('[test]\nb = "file"\n')
    monkeypatch.setenv("TEST_C", "env")

    # chdir *before* creating the CLI, b/c it will load the defaults immediately:
    subdir = tmp_path.joinpath("sub")
    subdir.mkdir()
    monkeypatch.chdir(subdir)

    result = Settings()  # Will be update by the CLI

    @cli_argparse.cli(Settings, default_loaders("test", [spath]))
    def cli(settings: Settings) -> Settings:
        return settings

    result = invoke(cli, "--d", "arg")
    assert result == Settings(
        a=subdir.joinpath("default"),
        b=spath.parent.joinpath("file"),
        c=subdir.joinpath("env"),
        d=subdir.joinpath("arg"),
    )


def test_multiple_invocations(invoke: Invoke) -> None:
    """
    A CLI function can be invoked multiple times w/o carrying state from call to call.
    """

    @settings
    class S:
        o: int = 0

    loaded_settings: List[S] = []

    @cli_argparse.cli(S, "example")
    def cli(settings: S) -> None:
        loaded_settings.append(settings)

    # The order of these invocations is important:
    invoke(cli, "--o=3")
    invoke(cli)
    assert loaded_settings == [S(3), S(0)]


@pytest.mark.skipif(
    not _compat.PY_310, reason='The output of arparse\'s "--help" has changed in 3.10'
)
def test_default_factory_multiple_invocations(
    invoke: Invoke, capsys: pytest.CaptureFixture
) -> None:
    """
    Default factories are not invoked by click when the CLI is generated.
    They are evaluate during the "convert" phase each time the CLI is invoked.
    """

    @settings
    class S:
        o: int = option(factory=lambda: len(loaded_settings) + 1)

    loaded_settings: List[S] = []

    @cli_argparse.cli(S, "example")
    def cli(settings: S) -> None:
        loaded_settings.append(settings)

    with pytest.raises(SystemExit):
        invoke(cli, "--help")
    out, err = capsys.readouterr()
    assert out == (
        "usage: cli [-h] [--o INT]\n"
        "\n"
        "options:\n"
        "  -h, --help  show this help message and exit\n"
        "\n"
        "S:\n"
        "  S options\n"
        "\n"
        "  --o INT     [default: (dynamic)]\n"
    )
    assert err == ""

    invoke(cli)
    invoke(cli)
    invoke(cli, "--o=100")
    assert loaded_settings == [S(1), S(2), S(100)]


def test_lazy_load_defaults(
    invoke: Invoke, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture
) -> None:
    """
    Defaults for CLIs are loaded lazily (when the CLI is invoked, not when it is
    constructed).

    On contrast to Click, the argparse CLI is only constructed when the CLI is invoked.
    """

    @settings
    class Settings:
        o1: int
        o2: int = 0

    loaded_settings: List[Settings] = []

    monkeypatch.setenv("EXAMPLE_O2", "1")

    @cli_argparse.cli(Settings, "example")
    def cli(settings: Settings) -> None:
        loaded_settings.append(settings)

    invoke(cli, "--o1=0")

    # Setting the env var to another value would have no effect w/o lazy loading:
    monkeypatch.setenv("EXAMPLE_O2", "2")
    invoke(cli, "--o1=0")

    assert loaded_settings == [Settings(0, 1), Settings(0, 2)]



"""
This module contains the settings loaders provided by Typed Settings and the
protocol specification that they must implement.
"""

import importlib.util
import logging
import os
from fnmatch import fnmatch
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Union,
    cast,
)

from ._compat import PY_311


if PY_311:
    import tomllib
else:
    import tomli as tomllib  # type: ignore[no-redef]

from . import cls_utils
from .dict_utils import set_path
from .exceptions import (
    ConfigFileLoadError,
    ConfigFileNotFoundError,
    InvalidOptionsError,
    UnknownFormatError,
)
from .types import (
    LoadedSettings,
    LoaderMeta,
    OptionInfo,
    OptionList,
    SettingsClass,
    SettingsDict,
)


__all__ = [
    "Loader",
    "DictLoader",
    "InstanceLoader",
    "EnvLoader",
    "FileLoader",
    "FileFormat",
    "PythonFormat",
    "TomlFormat",
    "OnePasswordLoader",
    "clean_settings",
]


LOGGER = logging.getLogger("typed_settings")


class Loader(Protocol):
    """
    **Protocol** that settings loaders must implement.

    Loaders must be callables (e.g., functions) with the specified signature.

    .. versionchanged:: 1.0.0
       Renamed ``load()`` to ``__call__()`` and also pass the settings
       class.
    """

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> Union[LoadedSettings, Iterable[LoadedSettings]]:
        """
        Load settings for the given options.

        Args:
            settings_cls: The base settings class for all options.
            options: The list of available settings.

        Return:
            A dict with the loaded settings.
        """
        ...


class FileFormat(Protocol):
    """
    **Protoco** that file format loaders for :class:`FileLoader` must
    implement.

    File format loaders must be callables (e.g., functions) with the specified
    signature.

    .. versionchanged:: 1.0.0
       Renamed ``load_file()`` to ``__call__()`` and also pass the settings
       class.
    """

    def __call__(
        self, path: Path, settings_cls: SettingsClass, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a given file and return them as a dict.

        Args:
            path: The path to the config file.
            settings_cls: The base settings class for all options.
            options: The list of available settings.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        ...


class _DefaultsLoader:
    """
    Return the default settings for a given settings class.

    Args:
        base_dir: Resolve relative paths in default settings relative to this directory.
            By default, use the user's current working directory.

            For tools like linters or test runners, you may want to use the root
            directory of the app that the tool operates in.
    """

    def __init__(self, base_dir: Path = Path()) -> None:
        self.base_dir = base_dir

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> LoadedSettings:
        settings: SettingsDict = {}

        # Populate dict with default settings.  This avoids problems with nested
        # settings classes for which no settings are loaded.
        for opt in options:
            if opt.has_no_default:
                continue
            if opt.default_is_factory:
                # Do not invoke default factories yet.  This should be done as late as
                # possible.  This is especially required for CLIs if you want to invoke
                # the same instance multiple times (e.g., in tests).
                continue
            set_path(settings, opt.path, opt.default)

        return LoadedSettings(settings, LoaderMeta(self, base_dir=self.base_dir))


class DictLoader:
    """
    Load settings from a dict of values.

    This is mainly for testing purposes.

    Args:
        settings: A (nested) dict of settings

    .. versionadded:: 23.1.0
    """

    def __init__(self, settings: dict) -> None:
        self.settings = settings

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> LoadedSettings:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        return LoadedSettings(self.settings, LoaderMeta(self))


class InstanceLoader:
    """
    Load settings from an instance of the settings class.

    Args:
        instance: The settings instance from which to load option values.

    .. versionadded:: 1.0.0
    """

    def __init__(self, instance: object) -> None:
        self.instance = instance

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> LoadedSettings:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        if not isinstance(self.instance, settings_cls):
            raise ValueError(
                f'"self.instance" is not an instance of {settings_cls}: '
                f"{type(self.instance)}"
            )
        cls_handler = cls_utils.find_handler(type(self.instance))
        return LoadedSettings(cls_handler.asdict(self.instance), LoaderMeta(self))


class EnvLoader:
    """
    Load settings from environment variables.

    Args:
        prefix: Prefix for environment variables, e.g., ``MYAPP_``.
        nested_delimiter: Delimiter for attribute names of nested classes.
    """

    def __init__(self, prefix: str, nested_delimiter: str = "_") -> None:
        self.prefix = prefix
        self.nested_delimiter = nested_delimiter

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> LoadedSettings:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        LOGGER.debug(f"Looking for env vars with prefix: {self.prefix}")

        env = os.environ
        values: SettingsDict = {}
        for o in options:
            varname = self.get_envvar(o)
            if varname in env:
                LOGGER.debug(f"Env var found: {varname}")
                set_path(values, o.path, env[varname])
            else:
                LOGGER.debug(f"Env var not found: {varname}")

        return LoadedSettings(values, LoaderMeta(self))

    def get_envvar(self, option: OptionInfo) -> str:
        """
        Return the envvar name for the he given option.
        """
        return f"{self.prefix}{option.path.upper().replace('.', self.nested_delimiter)}"


class FileLoader:
    """
    Load settings from config files.

    Settings of multiple files will be merged.  The last file has the highest
    precedence.  Files specified via an environment variable are loaded after
    the files passed to this class, i.e.:

    - First file from *files*
    - ...
    - Last file from *files*
    - First file from *env_var*
    - ...
    - Last file from *env_var*

    Mandatory files can be prefixed with ``!``.  Optional files will be ignored
    if they don't exist.

    Args:
        formats: A dict mapping glob patterns to :class:`FileFormat` instances.
        files: A list of filenames to try to load.
        env_var: Name of the environment variable that may hold additional file
            paths.  If it is ``None``, only files from *files* will be loaded.
    """

    def __init__(
        self,
        formats: Dict[str, FileFormat],
        files: Iterable[Union[str, Path]],
        env_var: Optional[str] = None,
    ) -> None:
        self.files = files
        self.env_var = env_var
        self.formats = formats

    def __call__(
        self, settings_cls: SettingsClass, options: OptionList
    ) -> List[LoadedSettings]:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.

        Raise:
            UnknownFormatError: When no :class:`FileFormat` is configured for a
                loaded file.
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
            InvalidOptionsError: If invalid settings have been found.
        """
        paths = self._get_config_filenames(self.files, self.env_var)
        loaded_settings: List[LoadedSettings] = []
        for path in paths:
            settings = self._load_file(path, settings_cls, options)
            meta = LoaderMeta(f"{type(self).__name__}[{path}]", base_dir=path.parent)
            loaded_settings.append(LoadedSettings(settings, meta))
        return loaded_settings

    def _load_file(
        self,
        path: Path,
        settings_cls: SettingsClass,
        options: OptionList,
    ) -> SettingsDict:
        """
        Load a file and return its cleaned contents.
        """
        # "clean_settings()" must be called for each loaded file individually
        # because of the "-"/"_" normalization.  This also allows us to tell
        # the user the exact file that contains errors.
        for pattern, ffloader in self.formats.items():
            if fnmatch(path.name, pattern):
                settings = ffloader(path, settings_cls, options)
                settings = clean_settings(settings, options, path)
                return settings

        raise UnknownFormatError(f"No loader configured for: {path}")

    @staticmethod
    def _get_config_filenames(
        files: Iterable[Union[str, Path]], env_var: Optional[str]
    ) -> List[Path]:
        """
        Concatenate *config_files* and files from env var *config_files_var*.
        """
        candidates = [(False, str(f)) for f in files]
        if env_var:
            LOGGER.debug(f"Env var for config files: {env_var}")
            candidates += [(True, fname) for fname in os.getenv(env_var, "").split(":")]
        else:
            LOGGER.debug("Env var for config files not set")

        paths = []
        for from_envvar, fname in candidates:
            _, flag, fname = fname.rpartition("!")
            if not fname:
                continue
            is_mandatory = flag == "!"
            try:
                path = Path(fname).resolve(strict=True)
            except FileNotFoundError:
                if is_mandatory:
                    LOGGER.error(f"Mandatory config file not found: {fname}")
                    raise
                if from_envvar:
                    LOGGER.warning(f"Config file from {env_var} not found: {fname}")
                else:
                    LOGGER.info(f"Config file not found: {fname}")
            else:
                LOGGER.debug(f"Loading settings from: {path}")
                paths.append(path)

        return paths


class PythonFormat:
    """
    Support for Python files.  Read settings from the given *section*.

    Args:
        section: The config file section to load settings from.
    """

    def __init__(
        self,
        cls_name: Optional[str],
        key_transformer: Callable[[str], str] = lambda k: k,
        flat: bool = False,
    ) -> None:
        self.cls_name = cls_name
        self.key_transformer = key_transformer
        self.flat = flat

    @staticmethod
    def to_lower(text: str) -> str:
        """
        Transform *text* to lower case.
        """
        return text.lower()

    def __call__(
        self, path: Path, settings_cls: SettingsClass, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a Python file and return them as a dict.

        Args:
            path: The path to the config file.
            options: The list of available settings.
            settings_cls: The base settings class for all options.  If ``None``, load
                directly from the module instead.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        module = self._import_module(path)

        def namespace2dict(namespace: object) -> SettingsDict:
            d = dict(namespace.__dict__)
            result = {}
            for k, v in d.items():
                if k.startswith("_"):
                    continue
                k = self.key_transformer(k)
                if isinstance(v, type):
                    v = namespace2dict(v)
                result[k] = v
            return result

        if self.cls_name is None:
            namespace = module
        else:
            try:
                namespace = getattr(module, self.cls_name)
            except AttributeError:
                return {}
        settings = namespace2dict(namespace)
        if self.flat:
            for o in options:
                key = f"{o.path.replace('.', '_')}"
                if key in settings:
                    val = settings.pop(key)
                    set_path(settings, o.path, val)
        return settings

    def _import_module(self, path: Path) -> object:
        module_name = path.stem
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:
            raise ConfigFileNotFoundError("No such file or directory: '{path}'")
        module = importlib.util.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)
        except SyntaxError as e:
            raise ConfigFileLoadError(str(e)) from e
        return module


class TomlFormat:
    """
    Support for TOML files.  Read settings from the given *section*.

    Args:
        section: The config file section to load settings from.
    """

    def __init__(self, section: Optional[str]) -> None:
        self.section = section

    def __call__(
        self, path: Path, settings_cls: SettingsClass, options: OptionList
    ) -> SettingsDict:
        """
        Load settings from a TOML file and return them as a dict.

        Args:
            path: The path to the config file.
            options: The list of available settings.
            settings_cls: The base settings class for all options.  If ``None``, load
                top level settings.

        Return:
            A dict with the loaded settings.

        Raise:
            ConfigFileNotFoundError: If *path* does not exist.
            ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        """
        try:
            with path.open("rb") as f:
                settings = tomllib.load(f)
        except FileNotFoundError as e:
            raise ConfigFileNotFoundError(str(e)) from e
        except (PermissionError, tomllib.TOMLDecodeError) as e:
            raise ConfigFileLoadError(str(e)) from e
        if self.section is not None:
            sections = self.section.split(".")
            for s in sections:
                try:
                    settings = settings[s]
                except KeyError:
                    return {}
        return cast(SettingsDict, settings)


class OnePasswordLoader:
    """
    Load settings from an item stored in a 1Password vault.

    You must must have installed and set up the `1Password CLI`_ in order
    for this loader to work.

    .. _1Password CLI: https://developer.1password.com/docs/cli/

    Args:
        item: The item to load
        vault: The vault in which to look for *item*.  By default, search all
            vaults.
    """

    def __init__(self, item: str, vault: Optional[str] = None) -> None:
        self.item = item
        self.vault = vault

        from . import _onepassword

        self._op = _onepassword

    def __call__(self, settings_cls: type, options: OptionList) -> LoadedSettings:
        """
        Load settings for the given options.

        Args:
            options: The list of available settings.
            settings_cls: The base settings class for all options.

        Return:
            A dict with the loaded settings.
        """
        option_names = [o.path for o in options]
        settings = self._op.get_item(self.item, self.vault)
        settings = {k: v for k, v in settings.items() if k in option_names}
        return LoadedSettings(settings, LoaderMeta(self))


def clean_settings(
    settings: SettingsDict, options: OptionList, source: Any
) -> SettingsDict:
    """
    Recursively check settings for invalid entries and raise an error.

    An error is not raised until all options have been checked.  It then lists
    all invalid options that have been found.

    Args:
        settings: The settings to be cleaned.
        options: The list of available settings.
        source: Source of the settings (e.g., path to a config file).
                It should have a useful string representation.

    Return:
        The cleaned settings.
    Raise:
        InvalidOptionsError: If invalid settings have been found.
    """
    invalid_paths = []
    valid_paths = {o.path for o in options}
    cleaned: SettingsDict = {}

    def _iter_dict(d: SettingsDict, prefix: str) -> None:
        for key, val in d.items():
            key = key.replace("-", "_")
            path = f"{prefix}{key}"

            if path in valid_paths:
                set_path(cleaned, path, val)
                continue

            if isinstance(val, dict):
                _iter_dict(val, f"{path}.")
            else:
                invalid_paths.append(path)

    _iter_dict(settings, "")

    if invalid_paths:
        joined_paths = ", ".join(sorted(invalid_paths))
        raise InvalidOptionsError(f"Invalid options found in {source}: {joined_paths}")

    return cleaned



from pathlib import Path
from typing import Iterable, Optional, Union


ROOT_DIR = Path().resolve().root


def find(
    filename: str,
    stop_dir: Union[str, Path] = ROOT_DIR,
    stop_files: Iterable[str] = (".git", ".hg"),
    *,
    start_dir: Optional[Path] = None,
) -> Path:
    """
    Search for a file in the current directory and its parents and return its
    path.

    If the file cannot be found until *stop_dir* is reached or a directory
    contains *stop_file*, return the input *filename*.

    Args:
        filename: The name of the file to find.
        stop_dir: Stop searching if the current search dir equals this one.
        stop_files: Stop searching if the current search dir contains
         this file.
        start_dir: Start directory for the search, defaults to the current directory.

    Returns:
        The resolved path to *filename* if found, else ``Path(filename)``.

    Examples:
        Find :file:`settings.toml` until the root of the current directory
        is reached::

            find("settings.py")

        Find :file:`pyproject.toml` only in the current Git project::

            find("pyproject.toml", stop_file=".git")

    .. versionchanged:: 23.1.0
       Added the *start_dir* parameter.
    """
    if start_dir is None:
        start_dir = Path.cwd()
    start = start_dir.joinpath(filename)
    for path in start.parents:
        p = path.joinpath(filename)
        if p.exists():
            return p

        if path == stop_dir:
            return start

        for stop_file in stop_files:
            if path.joinpath(stop_file).exists():
                return start

    return start



"""
Exceptions raised by Typed Settings.
"""


class TsError(Exception):
    """
    **Basse class** for all typed settings exceptions.
    """


class UnknownFormatError(TsError):
    """
    Raised when no file format loader is configured for a given config file.
    """


class ConfigFileNotFoundError(TsError):
    """
    Raised when a mandatory config file does not exist.
    """


class ConfigFileLoadError(TsError):
    """
    Raised when a config file exists but cannot be read or parsed/loaded.
    """


class InvalidOptionsError(TsError):
    """
    Raised when loaded settings contain an option that is not defined in the
    settings class.
    """


class InvalidValueError(TsError):
    """
    Raised the value of an option cannot be converted to the correct type.
    """


class InvalidSettingsError(TsError):
    """
    Raised when the loaded settings cannot be converted to an instances of the
    settings class.
    """



"""
A simple mypy plugin that makes the Typed Settings attrs aliases and wrappers
recognized by it.

You can activate the plugin via your :file:`pyproject.toml` or
:file:`mypy.ini`:

.. code-block:: toml

    # pyproject.toml
    [tool.mypy]
    plugins = ["typed_settings.mypy"]

.. code-block:: ini

    # mypy.ini
    [mypy]
    plugins=typed_settings.mypy
"""

try:
    from typing import Type

    from mypy.plugin import Plugin
    from mypy.plugins.attrs import attr_attrib_makers, attr_dataclass_makers
except ImportError:
    pass
else:
    # These work just like `attr.dataclass`.
    attr_dataclass_makers.add("attr.frozen")
    attr_dataclass_makers.add("typed_settings.cls_attrs.settings")

    # These are our `attr.ib` makers.
    attr_attrib_makers.add("attr.field")
    attr_attrib_makers.add("typed_settings.cls_attrs.option")
    attr_attrib_makers.add("typed_settings.cls_attrs.secret")

    class MyPlugin(Plugin):
        """
        Our plugin does nothing but it has to exist so this file gets loaded.
        """

        pass

    def plugin(version: str) -> Type[Plugin]:
        """
        Return the class for our plugin.
        """
        return MyPlugin



"""
Tests for "typed_settings.cls_attrs".
"""

import typing as t
from pathlib import Path

import attrs
import pytest

from typed_settings import cli_argparse, cli_click, constants
from typed_settings.cls_attrs import (
    SECRET,
    combine,
    evolve,
    option,
    secret,
    settings,
)


FieldFunc = t.Callable[..., t.Any]


@settings
class S:
    """A simple settings class with the TS aliases for attrs."""

    u: str = option()
    p: str = secret()


class TestFieldExtensions:
    """Tests for attrs field extensions."""

    @pytest.fixture
    def inst(self) -> S:
        """
        Return an instance of "S".
        """
        return S(u="spam", p="42")

    @pytest.fixture(params=[option, secret])
    def field_func(self, request: pytest.FixtureRequest) -> FieldFunc:
        """
        Generate two test params, one for "option", one for "secret".
        """
        return request.param

    def test_secret_repr_repr(self) -> None:
        """
        Secrets are represented by "***" and not printed directly.
        """
        assert str(SECRET) == "***"

    def test_secret_str(self, inst: S) -> None:
        """
        Values of secrets are obfuscated in the string repr.
        """
        assert str(inst) == "S(u='spam', p='*******')"

    def test_secret_repr_call(self, inst: S) -> None:
        """
        Values of secrets are obfuscated in the repr.
        """
        assert repr(inst) == "S(u='spam', p='*******')"

    def test_meta_not_set(self, field_func: FieldFunc) -> None:
        """
        The "help" and "click" entries are always present in the metadata,
        even if they are not explicitly set.
        """

        @settings
        class S:
            o: str = field_func()

        field = attrs.fields(S).o
        assert field.metadata == {
            constants.METADATA_KEY: {
                "help": None,
                cli_click.METADATA_KEY: {"help": None},
                cli_argparse.METADATA_KEY: {"help": None},
            },
        }

    def test_meta_help(self, field_func: FieldFunc) -> None:
        """
        "help" is stored directly in the meta and in the CLI options dicts.
        """

        @settings
        class S:
            o: str = field_func(help="spam")

        field = attrs.fields(S).o
        assert field.metadata == {
            constants.METADATA_KEY: {
                "help": "spam",
                cli_click.METADATA_KEY: {"help": "spam"},
                cli_argparse.METADATA_KEY: {"help": "spam"},
            },
        }

    def test_meta_help_override(self, field_func: FieldFunc) -> None:
        @settings
        class S:
            o: str = field_func(help="spam", click={"help": "eggs"})

        field = attrs.fields(S).o
        assert field.metadata == {
            constants.METADATA_KEY: {
                "help": "spam",
                cli_click.METADATA_KEY: {"help": "eggs"},
                cli_argparse.METADATA_KEY: {"help": "spam"},
            },
        }

    def test_meta_click_params(self, field_func: FieldFunc) -> None:
        """
        "help" can be overwritten via "click" options.
        """

        @settings
        class S:
            o: str = field_func(click={"param_decls": ("-o",)})

        field = attrs.fields(S).o
        assert field.metadata == {
            constants.METADATA_KEY: {
                "help": None,
                cli_click.METADATA_KEY: {"help": None, "param_decls": ("-o",)},
                cli_argparse.METADATA_KEY: {"help": None},
            },
        }

    def test_meta_merge(self, field_func: FieldFunc) -> None:
        """
        If metadata is already present, it is not overridden.
        """

        @settings
        class S:
            o: str = field_func(
                metadata={"spam": "eggs"},
                help="halp!",
                click={"param_decls": ("-o",)},
            )

        field = attrs.fields(S).o
        assert field.metadata == {
            "spam": "eggs",
            constants.METADATA_KEY: {
                "help": "halp!",
                cli_click.METADATA_KEY: {"help": "halp!", "param_decls": ("-o",)},
                cli_argparse.METADATA_KEY: {"help": "halp!"},
            },
        }


class TestEvolve:
    """
    Tests for `evolve`.

    Copied from attrs and adjusted/reduced.
    """

    def test_validator_failure(self) -> None:
        """
        TypeError isn't swallowed when validation fails within evolve.
        """

        @settings
        class C:
            a: int = option(validator=attrs.validators.instance_of(int))

        with pytest.raises(TypeError) as e:
            evolve(C(a=1), a="some string")
        m = e.value.args[0]

        assert m.startswith("'a' must be <class 'int'>")

    def test_private(self) -> None:
        """
        evolve() acts as `__init__` with regards to private attributes.
        """

        @settings
        class C:
            _a: str

        assert evolve(C(1), a=2)._a == 2  # type: ignore

        with pytest.raises(TypeError):
            evolve(C(1), _a=2)  # type: ignore[arg-type]

        with pytest.raises(TypeError):
            evolve(C(1), a=3, _a=2)  # type: ignore[arg-type]

    def test_non_init_attrs(self) -> None:
        """
        evolve() handles `init=False` attributes.
        """

        @settings
        class C:
            a: str
            b: int = option(init=False, default=0)

        assert evolve(C(1), a=2).a == 2  # type: ignore

    def test_regression_attrs_classes(self) -> None:
        """
        evolve() can evolve fields that are instances of attrs classes.

        Regression test for #804
        """

        @settings
        class Child:
            param2: str

        @settings
        class Parent:
            param1: Child

        obj2a = Child(param2="a")
        obj2b = Child(param2="b")

        obj1a = Parent(param1=obj2a)

        assert Parent(param1=Child(param2="b")) == evolve(obj1a, param1=obj2b)

    def test_recursive(self) -> None:
        """
        evolve() recursively evolves nested attrs classes when a dict is
        passed for an attribute.
        """

        @settings
        class N2:
            e: int

        @settings
        class N1:
            c: N2
            d: int

        @settings
        class C:
            a: N1
            b: int

        c1 = C(N1(N2(1), 2), 3)
        c2 = evolve(c1, a={"c": {"e": 23}}, b=42)

        assert c2 == C(N1(N2(23), 2), 42)

    def test_recursive_attrs_classes(self) -> None:
        """
        evolve() can evolve fields that are instances of attrs classes.
        """

        @settings
        class Child:
            param2: str

        @settings
        class Parent:
            param1: Child

        obj2a = Child(param2="a")
        obj2b = Child(param2="b")

        obj1a = Parent(param1=obj2a)

        result = evolve(obj1a, param1=obj2b)
        assert result.param1 is obj2b  # type: ignore


class TestCombine:
    """
    Tests for "combine()".
    """

    def test_combine(self) -> None:
        """
        A base class and nested classes can be combined into a single, composed
        class.
        """

        @attrs.define
        class Nested1:
            a: str = ""

        @attrs.define
        class Nested2:
            a: str = ""

        # Dynamic composition
        @attrs.define
        class BaseSettings:
            a: str = ""

        Composed = combine(
            "Composed",
            BaseSettings,
            {"n1": Nested1(), "n2": Nested2()},
        )
        assert Composed.__name__ == "Composed"
        assert [(f.name, f.type, f.default) for f in attrs.fields(Composed)] == [  # type: ignore
            ("a", str, ""),
            ("n1", Nested1, Nested1()),
            ("n2", Nested2, Nested2()),
        ]

    def test_duplicate_attrib(self) -> None:
        """
        Raise an error if a nested class placed with attrib name that is
        already used by the base class.
        """

        @attrs.define
        class Nested1:
            a: str = ""

        # Dynamic composition
        @attrs.define
        class BaseSettings:
            a: str = ""

        with pytest.raises(ValueError, match="Duplicate attribute for nested class: a"):
            combine(
                "Composed",
                BaseSettings,
                {"a": Nested1()},
            )

    def test_docstring(self) -> None:
        """
        The created class copies the costring from the base class.
        """

        @attrs.define
        class Nested1:
            a: str = ""

        # Dynamic composition
        @attrs.define
        class BaseSettings:
            """Le doc string."""

            a: str = ""

        Composed = combine("Composed", BaseSettings, {"n1": Nested1()})
        assert Composed.__doc__ == "Le doc string."

    def test_postponed_annotations(self) -> None:
        """
        The created class copies the annotations from the base and adds entries for all
        nested classes.

        See: https://gitlab.com/sscherfke/typed-settings/-/issues/54
        """

        @attrs.define
        class Nested1:
            a: "str" = ""

        @attrs.define
        class BaseSettings:
            a: "bool" = False
            p: "Path" = Path.cwd()

        Composed = combine("Composed", BaseSettings, {"n1": Nested1()})

        # Composed has __annotations__ populated with base and nested attribs
        assert Composed.__annotations__ == {
            "a": "bool",
            "p": "Path",
            "n1": Nested1,
        }

        # Types can be resolved
        Composed = attrs.resolve_types(Composed, globalns=globals())
        fields = attrs.fields(Composed)
        assert fields.a.type is bool
        assert fields.p.type is Path
        assert fields.n1.type is Nested1
